
require "druby/profile/collector"
require "druby/profile/require"

module DRuby
  class Profile

    def initialize(filename, features)
      collector = Collector.new(filename)

      klasses = features.map do |name|
        # for each component foo, we do the following:
        # require "druby/profile/foo"
        # DRuby::Profile::Foo(collector)
        require "druby/profile/#{name}"
        klass = DRuby::Profile.const_get name.capitalize
        klass.new(collector)
      end

      DRuby::Profile::Interceptor.enabled = true
      klasses.each {|k| k.enable}

      at_exit do
        DRuby::Profile::Interceptor.enabled = false
        klasses.each {|k| k.disable}
        collector.finalize
      end
    end

    # checks if recv#meth_name is a method of klass
    def self.valid_recv?(klass, recv, meth_name) 
      meth = recv.method meth_name
      mod = nil
      case meth.inspect
      when /Method: ([^#]+)\((.*)\)#(.+)>$/
        cls = $2
      when /#<Method: #<[^>]*>\((.*)\)#(.+)>$/ # Anonymous class object
        cls = $1
      when /^#<UnboundMethod: ([^#]+)#(.+)>$/
        return false
      when /^#<Method: ([^.]+)\((.*)\)\.(.+)>$/
        cls = $2 # XXX
      when /^#<Method: ([^.]+)\.(.+)>$/ # singleton method
        return false
      when /^#<Method: ([^#\(]+)#(.+)>$/
        cls = $1
      else 
        fail "[FATAL] other #{meth.inspect}"
      end
      mod = eval(cls) 
      mod == klass 
    end

  end
end

