
open Cfg

(** The signature for a CFG pretty printer.  Where appropraite, type
    declaration present in the Cfg module have a corresponding format
    function here.  The naming scheme is meant to exactly match that
    of the type signatures.  Thus if there is a type foo_bar, then the
    corresponding formatting function is named format_foo_bar.  Note
    that some intermediate/helper types in Cfg do not have their own
    formatting function, and this signature includes a few helpers of
    its own, which have no corresponding Cfg types. 
 *)
module type CfgPrinter = sig
  val format_unary_op : Format.formatter -> unary_op -> unit
  val format_binary_op : Format.formatter -> binary_op -> unit
  val format_identifier : Format.formatter -> identifier -> unit
  val format_msg_id : Format.formatter -> msg_id -> unit
  val format_star_expr : Format.formatter -> star_expr -> unit
  val format_expr : Format.formatter -> expr -> unit
  val format_literal : Format.formatter -> literal -> unit
  val format_tuple_expr : Format.formatter -> tuple_expr -> unit
  val format_lhs : Format.formatter -> lhs -> unit
  val format_def_name : Format.formatter -> def_name -> unit
  val format_method_arg_expr : Format.formatter -> method_arg_expr -> unit
  val format_any_formal : Format.formatter -> any_formal -> unit
  val format_class_kind : Format.formatter -> class_kind -> unit

  val format_method_call : Format.formatter -> method_call -> unit
  val format_case : Format.formatter -> case_block -> unit
  val format_codeblock : Format.formatter -> codeblock -> unit
  val format_rescue_guard : Format.formatter -> rescue_guard -> unit 

  val format_stmt : Format.formatter -> stmt -> unit
  val format_cfg : Format.formatter -> stmt -> unit

  (* helpers*) 
  val format_def_id : Format.formatter -> identifier -> unit
  val format_formal_tuple : Format.formatter -> any_formal list -> unit
  val format_formals : Format.formatter -> any_formal list -> unit
  val format_target_and_msg : Format.formatter -> method_target_expr * msg_id -> unit

  val print_cfg : stmt -> unit
  val string_of_expr : expr -> string
  val string_of_cfg : stmt -> string
  val print_stmt : out_channel -> stmt -> unit
end

(** This module formats CFGs so that they are valid Ruby code. *)
module CodePrinter : CfgPrinter

(** This module shares most of the implementation of [CodePrinter],
    but simplifies away some of the translations done such as
    temporary variables and implicit returns.  Code formatted by this
    module is, in general, NOT valid Ruby code.
*)
module ErrorPrinter : CfgPrinter
