
open Printf
open Dynamic
open Yaml
open Utils
open Dyn_builder

module Domain = YString
module CoDomain = YString

let name = "reflection"

let instrument_ast e = e
let transform_ast t e = e 

open Cfg
module C = Cfg.Abbr

let refl_watch meth = 
  {src_msg = `ID_MethodName meth;
   dst_target = C.access_path ["DRuby";"Profile";"Reflection"];
   dst_msg = `ID_MethodName "watch";
  }

let instrument_cfg s = 
  let trans = [
    refl_watch "instance_variable_get";
    refl_watch "class_variable_get";
    refl_watch "const_get";
    refl_watch "instance_variable_set";
    refl_watch "class_variable_set";
    refl_watch "const_set";
    (*
      refl_watch "attr";
      refl_watch "attr_accessor";
      refl_watch "attr_writer";
      refl_watch "attr_reader";
    *)
  ] in Cfg.visit_stmt (new watch_visitor trans) s

module Base = struct
  module Domain = Domain
  module Codomain = CoDomain
  let preamble = []
  let when_guard str = C.str str
end
            
class refl_visitor yaml_map = 
object(self)
  inherit default_visitor
  method visit_stmt s = match s.snode with
    | MethodCall(lhs,mc) -> begin match mc with
        | {mc_msg=`ID_MethodName ("instance_variable_get"|"class_variable_get" as get);
           mc_args=(#expr as arg)::_} ->
            let module M = struct
              include Base
              let case_guard = arg
              let else_body = dead_method get [(arg :> method_arg_expr)] s.pos
              let when_body str = 
                let body = C.next ~v:(C.var str) s.pos in
                let cb = {cb_formals=[];cb_body=body} in
                  C.mcall ?lhs ~targ:mc.mc_target (`ID_MethodName "instance_eval") [] ~cb s.pos
            end in let module T = CaseTransformer(M) in
              T.transform yaml_map get s.pos

        | {mc_msg=`ID_MethodName ("instance_variable_set"|"class_variable_set" as set);
           mc_args=[(#expr as arg);(#star_expr as rhs)]} ->
            let module M = struct
              include Base
              let case_guard = arg
              let else_body = dead_method set [(arg :> method_arg_expr)] s.pos
              let when_body str = 
                let body = C.seq [C.assign (C.var str) (rhs:>tuple_expr) s.pos;
                                  C.next ~v:(C.var str) s.pos] s.pos 
                in
                let cb = {cb_formals=[];cb_body=body} in
                  C.mcall ?lhs ~targ:mc.mc_target (`ID_MethodName "instance_eval") [] ~cb s.pos
            end in let module T = CaseTransformer(M) in
              T.transform yaml_map set s.pos

        | {mc_msg=`ID_MethodName ("const_get"|"const_set"); mc_args=(`Lit_Atom _)::_} ->
            (* already uses a literal, no need to transform *)
            Visitor.SkipChildren

        | {mc_msg=`ID_MethodName ("const_set"|"const_get" as const);mc_args=(#expr as arg)::rest} ->
            let tmp = C.local "__druby_profiled_const" in
            let to_s = C.mcall ~lhs:tmp ~targ:arg (`ID_MethodName "to_s") [] s.pos in
            let module M = struct
              include Base
              let preamble = [to_s]
              let case_guard = arg
              let else_body = dead_method const [(arg :> method_arg_expr)] s.pos
              let when_body str = 
                mkstmt (MethodCall(lhs,{mc with mc_args=(C.atom str)::rest})) s.pos
            end in let module T = CaseTransformer(M) in
              T.transform yaml_map const s.pos

        | _ -> Visitor.DoChildren
      end


    (*
      | MethodCall(lhs,{mc_target=targ;mc_args=arg::rest; mc_cb=None;
      mc_msg=`ID_MethodName ("attr" as attr)}) ->
      build_attr_method t lhs targ attr arg rest s.pos

      | MethodCall(lhs,{mc_target=targ;mc_args=[arg]; mc_cb=None;
      mc_msg=`ID_MethodName ("attr_accessor" as attr)}) ->
      build_attr_method t lhs targ attr arg [] s.pos

      | MethodCall(lhs,{mc_target=targ;mc_args=[arg]; mc_cb=None;
      mc_msg=`ID_MethodName ("attr_reader" as attr)}) ->
      build_attr_method t lhs targ attr arg [] s.pos

      | MethodCall(lhs,{mc_target=targ;mc_args=[arg]; mc_cb=None;
      mc_msg=`ID_MethodName ("attr_writer" as attr)}) ->
      build_attr_method t lhs targ attr arg [] s.pos
    *)
    | _ -> Visitor.DoChildren
end

(*

let build_attr_method yaml_map lhs targ attr arg extra_args pos = 
  let when_ str = 
    let var_string = C.str str in
    let m = C.mcall ?lhs ~targ (`ID_MethodName attr) (var_string::extra_args) pos in
      var_string, m
  in
  let whens = List.map when_ (yaml_map attr pos) in
  let default = dead_method attr [(arg:>method_arg_expr)] pos in
  let s' = C.case arg whens ~default pos in
    Visitor.ChangeTo s'
*)
let transform_cfg t s = visit_stmt (new refl_visitor t) s
