open Ast

type unary_op = Ast.unary_op

type method_storage =
  | Class_Method
  | Instance_Method
  | Module_Method

type formal_arg = {
  fa_name : string;
  fa_star : bool;
  fa_amp  : bool;
  fa_def  : expression option;
}

and ruby_method = {
  meth_name: string;
  meth_storage: method_storage;
  meth_params : formal_arg list;
  meth_body: statement list;
}

and ruby_class = {
  class_name : string;
  class_super : ruby_class option;
  class_mixins : ruby_module list;
  class_methods : ruby_method list;
  instance_methods : ruby_method list;
}

and ruby_module = {
  module_name : string;
}

and when_clause = {
  when_exprs: expression list;
  when_body: block;
}

and case_block = {
  case_expr : expression;
  when_clauses: when_clause;
  else_clause: block;
}

and rescue_block = {
  rescue_exceptions: string list;
  rescue_body: block;
}

and exn_block = {
  try_block : block;
  rescue_blocks: rescue_block list;
  else_block: block;
  ensure_block: block;
}

and block = {
  block_exprs: expression list;
}

and statement = 
  | Alias of expression * expression

  | BeginBlock of expression list
  | EndBlock of expression list

  | Module of ruby_module

  | Class of ruby_class

  | ExnBlock of exn_block

  | Case of case_block

  | Loop of expression * statement list
  | Conditional of expression * statement list * statement list

  | Block of block

  | Break
  | Redo
  | Next
  | Yield
  | Retry


and expression = 
  | Literal of literal
  | Identifier of id_kind * string

  | UnaryOp of unary_op * expression
  | BinaryOp of expression * binary_op * expression

  | AccessPath of expression list
  | MethodCall of expression * expression list * expression option

  | Return of expression option

and literal = 
  | Lit_FixNum of int
  | Lit_BigNum of Big_int.big_int
  | Lit_Float of float
  | Lit_String of string
  | Lit_Atom of string
  | Lit_Regexp of string * string
  | Lit_True
  | Lit_False
  | Lit_Nil
  | Lit_Hash of expression list
  | Lit_Array of expression list

  | Lit_Undef
  | Lit_Super
  | Lit_Self

(*
let rec to_expression expr = match expr with
  | E_If(guard,true_branch,false_branch) ->
      Conditional(to_expression guard,
		 to_stmt_list true_branch,
		 to_stmt_list false_branch)
      
  | E_Unless(guard,false_branch,true_branch) ->
      Conditional(UnaryOp(Op_UNot,to_expression guard),
		 to_stmt_list true_branch,
		 to_stmt_list false_branch)
  | _ -> failwith "fixme"
    
and to_stmt_list lst = match lst with
  | E_Constant _ -> []
  | _ -> []

*)
