
%{
  open Ast
    
  let rec add_eq = function
    | E_Identifier(k,s) -> E_Identifier(ID_Assign(k), s)
    | E_Binop(e1,Op_SCOPE,e2) -> E_Binop(e1,Op_SCOPE, add_eq e2)
    | E_Binop(e1,Op_DOT,e2) -> E_Binop(e1,Op_DOT, add_eq e2)
    | E_Operator(Op_AREF) -> E_Operator(Op_ASET)
    | _ -> failwith "add_eq"
	
  let make_class_var = function
    | E_Identifier(k,s) -> E_Identifier(ID_Class,"@@" ^ s)
    | _ -> failwith "make_class_var"

  let make_inst_var = function
    | E_Identifier(k,s) -> E_Identifier(ID_Instance,"@" ^ s)
    | _ -> failwith "make_inst_var"

  let rec flatten_commas = function
    | E_Binop(lhs,Op_COMMA,rhs) -> lhs :: (flatten_commas rhs)
    | x -> [x]
%}

%token T_DOT            /* . */
%token T_COMMA          /* , */
%token T_AT             /* @ */
%token T_PLUS 		/* + */
%token T_UPLUS 		/* unary + */
%token T_MINUS		/* - */
%token T_UMINUS		/* unary - */
%token T_POW		/* ** */
%token T_CMP  		/* <=> */
%token T_ASSIGN         /* = */
%token T_EQ  		/* == */
%token T_EQQ  		/* === */
%token T_NEQ  		/* != */
%token T_GEQ  		/* >= */
%token T_LEQ  		/* <= */
%token T_LT 		/* < */
%token T_GT  		/* > */
%token T_ANDOP T_OROP	/* && and || */
%token T_MATCH T_NMATCH	/* =~ and !~ */
%token T_DOT2 T_DOT3	/* .. and ... */
/*%token T_ASET*/	        /* []= */
%token T_LSHFT T_RSHFT	/* << and >> */
%token <string> T_OP_ASGN	/* +=, -=  etc. */
%token T_ASSOC		/* => */
%token T_LPAREN		/* ( */
%token T_LPAREN_ARG	/* ( */
%token T_RPAREN		/* ) */
%token T_LBRACK_ARG	/* [ as in x[y] = 2 */
%token T_LBRACK		/* [ as in x = [] */
%token T_RBRACK		/* ] */
%token T_LBRACE		/* { */
%token T_RBRACE		/* } */
%token T_STAR		/* * */
%token T_USTAR		/* * */
%token T_SLASH		/* / */
%token T_PERCENT	/* % */
%token T_QUESTION       /* ? */
%token T_COLON          /* : */
%token T_CARROT         /* ^ */
%token T_VBAR           /* | */
%token T_BANG           /* ! */
%token T_TILDE          /* ~ */
%token T_AMPER		/* & */
%token T_UAMPER		/* & */
%token T_SEMICOLON      /* ; */
%token T_SCOPE          /* x::y */
%token T_USCOPE      /* f ::y */

%token <string> T_UID
%token <string> T_LID
%token <string> T_GLOBAL_VAR
(*%token <string> T_INST_VAR
%token <string> T_CLASS_VAR*)
%token <string> T_ATOM
/*%token <string list> T_SCOPED_ID*/
%token <int> T_FIXNUM
%token <Big_int.big_int> T_BIGNUM
%token <float> T_FLOAT
%token <string*string> T_REGEXP
%token <string> T_TICK
%token <string> T_BUILTIN_VAR

/* keywords */
%token  K_CLASS
	K_MODULE
	K_DEF
	K_END
	K_ALIAS
	K_BEGIN
	K_RESCUE
	K_ENSURE
	K_IF
	K_UNLESS
	K_THEN
	K_ELSIF
	K_ELSE
	K_CASE
	K_WHEN
	K_WHILE
	K_UNTIL
	K_FOR
	K_IN
	K_DO
	K_RETURN
	K_AND
	K_OR
	K_NOT
	K_DEFINED
	K_lBEGIN
	K_lEND
	K_NIL
	K_YIELD
/*
	K_SELF
	K_SUPER
	K_UNDEF
	K_BREAK
	K_REDO
	K_TRUE
	K_FALSE
	K_RETRY
	K_NEXT*/
     /* K___LINE__
	K___FILE__
      */

%token <string> T_SINGLE_STRING /* 'string' */
%token <string> T_DOUBLE_STRING /* "string" */
%token <string*string> T_USER_STRING /* %!string! */
%token <int> T_HEREDOC /* <<DELIM ... DELIM */
%token T_EOL
%token T_EOF

%nonassoc K_IF K_UNLESS K_UNTIL K_WHILE K_RESCUE


%nonassoc /*T_USTAR*/ T_UAMPER
%left  K_OR K_AND
%right K_NOT
%nonassoc K_DEFINED
%right T_ASSIGN T_OP_ASGN

%nonassoc T_ASSOC

%right T_QUESTION T_COLON
%nonassoc T_DOT2 T_DOT3
%left  T_OROP
%left  T_ANDOP
%nonassoc  T_CMP T_EQ T_EQQ T_NEQ T_MATCH T_NMATCH
%left  T_GT T_GEQ T_LT T_LEQ
%left  T_VBAR T_CARROT
%left  T_AMPER
%left  T_LSHFT T_RSHFT
%left  T_PLUS
%left  T_MINUS
%left  T_STAR T_SLASH T_PERCENT
%right T_UMINUS
%right T_POW
%right T_BANG T_TILDE T_UPLUS
(*%left T_SCOPE
%right T_USCOPE*)

%start <Ast.expr list> main

%%

n:
  | list(T_EOL) {}

main:
  | n; program = expression_list; T_EOF { program }

expression_list:
  | {[]}
  | T_SEMICOLON; n; es = expression_list { es }
  | e = expression { [e] }
  | e = expression; T_SEMICOLON; n; es=expression_list
  | e = expression; nonempty_list(T_EOL); es = expression_list
     { e:: es}

expression:
     (* the standard expressions go here, this is also the production that can
        occur inside of conditional guards
     *)
  | e = guard_expr { e }

  (* statements really.  These evaluate to nil and so don't respond to any
     (reasonable) messages, therefore we won't syntacally allow that *)
  | e = non_meth_expr { e }

  (* these modifiers can not be nested so they must appear at the top level *)
  | e = expression; K_IF; n;guard = expression     { E_If(guard, [e], []) }
  | e = expression; K_UNLESS;n; guard = expression { E_Unless(guard, [e], []) }
  | e = expression; K_UNTIL;n; guard = expression  { E_Until(guard, [e]) }
  | e = expression; K_WHILE;n; guard = expression  { E_While(guard, [e]) }

  (* "rescue" can, apparently, be a modifier too *)
    (* used to be compound_expr *)
  | e = expression; K_RESCUE;n; expression;
    { e }

guard_expr: e = guarded(tuple_expr) { e }
inner_guard: e = guarded(compound_with_assign) { e }

    (* top level expressions that can be also be a conditional guard *)
guarded(sub):
  | e1 = guarded(sub); K_AND; n; e2 = guarded(sub)
     { E_Binop(e1,Op_kAND,e2) }
  | e1 = guarded(sub); K_OR; n; e2 = guarded(sub)
    { E_Binop(e1,Op_kOR,e2) }
  | e = sub { e }

(* this is where it starts to get ugly.  There are several syntactic quarks
   First, an object can have arguments without ()'s, like:
     o a1, a2, a3
   but this obviously doesn't nest.  We also need to handle multiple assign 
   specially 
   x,(y,z)) = a,(b,c)
*)
tuple_expr:
    (* method invocation without parens *)
  | e = outer_method_call { e }
  | T_BANG; e = outer_method_call { E_Unary(Op_UBang,e) }

  | e = obj_or_paren; T_ASSIGN;n; e2 = tuple_expr
    { E_Binop(e,Op_ASSIGN,e2) }

  | e = compound_expr { e }

    (* certain places allow tuples, such as nested inside a multiple 
       assignment (see next production) *)
  | e=obj_or_paren; T_COMMA;n;
    lst = separated_nonempty_list(pair(T_COMMA,n),obj_or_paren);
    { E_Tuple(e::lst) }

    (* handle multiple assignment.  Ruby does not allow these to be nested, 
         x,y = z,w = 1,2 
       doesn't work like you think it should.  We try to mirror that
       behavior here 
    *)
(*  | e = obj_or_paren; T_COMMA; n; lst = comma_list_suffix_empty(obj_or_paren); *)

  | e = obj_or_paren; T_COMMA;n; es = mlhs; T_ASSIGN; n;
    rhs = separated_nonempty_list(pair(T_COMMA,n),compound_with_assign);
    {let t1 = match es with [] -> e | _ -> E_Tuple(e::es) in
     let t2 = match rhs with hd::[] -> hd | _ -> E_Tuple(rhs) in
       E_Binop(t1,Op_ASSIGN,t2)}

mlhs:
  | { [] }
  | T_USTAR { [E_UOperator Op_UStar]}
  | e = obj_or_paren; {[e]}
  | e = obj_or_paren; T_COMMA;n; rst = mlhs; {e::rst}

outer_method_call:
  | e = outer_base; arg = return_arg
    { match e with
	| E_Yield [] -> E_Yield [arg]
	| E_Return [] -> E_Return [arg]
	| _ -> E_MethodCall(e,[arg], None ) }

  | e = outer_base; arg1 = return_arg;pair(T_COMMA,n);
    rest = separated_nonempty_list(pair(T_COMMA,n),return_arg_paren);
    { let args = arg1::rest in match e with
	| E_Yield [] -> E_Yield args
	| E_Return [] -> E_Return args
	| _ -> E_MethodCall(e,args, None ) }

outer_base:
  | e = obj_expr {e}

obj_or_paren:
  | o = obj_expr { o }
  | e = star_exp(obj_expr)   { e }
  | e = paren_exp { e }
  | e = paren_brackets { e }

paren_brackets:
  | e = paren_exp; 
    T_LBRACK_ARG; args = separated_nonempty_list(T_COMMA,inner_guard); T_RBRACK;
    { E_MethodCall(
	E_Binop(e,Op_DOT,E_Operator Op_AREF),
	args, None)}

(* top level expressions (statements really) that evalutate to nil *)
non_meth_expr:
  | K_ALIAS; e1 = meth_or_atom; e2 = meth_or_atom { E_Alias(e1,e2) }
  | K_BEGIN; n; T_LBRACE;n; body = expression_list; T_RBRACE {E_BeginBlock body}
  | K_END; n; T_LBRACE;n; body = expression_list; T_RBRACE {E_EndBlock body}

ctrl_sep(keyword):
    | T_SEMICOLON; n; 
    | T_COLON; n;
    | nonempty_list(T_EOL)
    | n; keyword; n {}

class_inheritance:
 | {None}
 | T_LT; id = obj_expr { Some (Class_Inherit id) }

body_exn:
 | body = expression_list;
   resc_e = list(rescue_clause); 
   ens_e = ensure_clause; else_e=case_else;
   { {body_exprs = body;
      rescue_exprs = resc_e;
      ensure_expr = ens_e;
      else_expr = else_e} }

(* An expression that can accept a message *)
obj_expr:
  | K_MODULE; name=scoped(identifier); n; body=body_exn; K_lEND 
      { E_ModuleDef (name,body) }

  | K_CLASS; n; class_name = scoped(identifier); inh = class_inheritance;
    n; body = body_exn; K_lEND
    { E_ClassDef(class_name, inh, body) }

  | K_CLASS; T_LSHFT; id = method_name; n; body = body_exn; K_lEND
    { E_ClassDef(E_Identifier(ID_Uppercase, ""), Some (Inst_Inherit id), body) }

  | K_lBEGIN; n; body = body_exn;  K_lEND
    { E_ExnBlock body }

  | K_DEF; meth_name=method_name; formals=method_formals;
    body = body_exn;K_lEND
    { E_MethodDef (meth_name, formals, body) }

  | K_FOR; vars=comma_list_suffix(formal_arg); K_IN; 
    range=guarded(method_arg);ctrl_sep(K_DO); body=expression_list; K_lEND
    { E_For(vars,range,body) }

  | K_CASE; nonempty_list(T_EOL); whens=when_clauses; else_e=case_else; K_lEND;
    { E_Case {case_guard = E_Empty; case_whens = whens; case_else = else_e} }

  | K_CASE;n; e=guard_expr; n; whens=when_clauses; else_e=case_else; K_lEND;
    { E_Case {case_guard = e; case_whens = whens; case_else = else_e} }

  | T_LBRACK; n; T_RBRACK { E_Array [] }
  | T_LBRACK; n; a = array_body; T_RBRACK;  { E_Array a }

  | K_RETURN { E_Return[] }
  | K_RETURN; T_LPAREN_ARG;n;e = separated_list(T_COMMA,inner_guard);T_RPAREN
    { E_Return e }

  | e = constant   { e }

  | e = base_obj { e }
  | e = base_obj; cb = code_block {E_MethodCall(e,[],Some cb)}

  | e = base_obj; lparen(*T_LPAREN_ARG*); n;
    args = separated_nonempty_list(pair(T_COMMA,n),inner_guard); n;T_RPAREN;
    cb = option(code_block);
    { match e with 
	| E_Yield [] -> E_Yield args
	| _ -> E_MethodCall(e,args,cb) }

  | e = base_obj; lparen (*T_LPAREN_ARG*); n; T_RPAREN; cb = option(code_block);
    { match e with 
	| E_Yield [] -> E_Yield []
	| _ -> E_MethodCall(e,[],cb) }

  | e = method_with_codeblock(obj_expr) { e }

base_obj:
  | K_DEFINED {E_UOperator(Op_UDefined)}
  | K_YIELD { E_Yield[] }
  | e = identifier { e }
  | T_USCOPE; id = identifier;
  | T_USCOPE; id = keyword_as_id;
    { E_Unary(Op_UScope,id)}
  | id1 = obj_expr; T_SCOPE; id2 = identifier;
  | id1 = obj_expr; T_SCOPE; id2 = keyword_as_id;
    { E_Binop(id1,Op_SCOPE,id2) }

(*TODO inner guard? *)
array_body: 
  | e = outer_method_call { [e] }
  | e = comma_list_suffix(compound_with_assign) { e }

hash_body: e = comma_list_suffix(compound_with_assign) { e }
comma_list_suffix_empty(base):
  | { [] }
  | e = comma_list_suffix(base) {e}
comma_list_suffix(base):
  | e = base; n; { [e] }
  | e = base; T_COMMA; n; es = comma_list_suffix(base)  { e :: es }
  | e = base; T_COMMA n; { [e] }

paren_exp:
(*  | T_LPAREN; T_RPAREN;    { E_Literal(Lit_Nil) }*)
  | T_LPAREN_ARG; n; es = expression_list; T_RPAREN; 
  | T_LPAREN; n; es = expression_list; T_RPAREN; 
    { match es with
	| [] -> E_Literal(Lit_Nil)
	| hd::[] -> hd
	| l -> E_Block l }

%inline star_exp(rec_expr):
  | T_USTAR; o = rec_expr { E_Unary(Op_UStar,o) }

method_with_codeblock(base):
  | e = method_call(base) { e }
  | m = method_call(base); cb = code_block;
    {match m with 
       | E_MethodCall(x,l, None) ->  E_MethodCall(x,l,Some cb)
       | _ -> E_MethodCall(m,[], Some cb) 
    }

method_call(base):
  (* o.m and o.m() *)
  | e = base; T_DOT; n;m = message_identifier 
  | e = base; T_DOT; n;m = message_identifier; lparen(*T_LPAREN_ARG*); n; T_RPAREN;
    { E_MethodCall(E_Binop(e,Op_DOT,m),[],None) }
  (* o.m(e,...e) *)
  | e = base; T_DOT; n;m = message_identifier; lparen(*T_LPAREN_ARG*); n; 
    args = separated_nonempty_list(pair(T_COMMA,n),inner_guard);n;
    T_RPAREN;
    { E_MethodCall(E_Binop(e,Op_DOT,m),args,None) }

  (* o[e,e] *)
  | e = base; 
    T_LBRACK_ARG; args = separated_nonempty_list(T_COMMA,inner_guard); T_RBRACK;
    { E_MethodCall(
	E_Binop(e,Op_DOT,E_Operator Op_AREF),
	args, None)}
    
message_identifier:
  | e = assignable { e }
  | op = bin_op_no_assign { E_Operator op }
  | T_UPLUS       { E_Operator Op_PLUS }
  | T_UMINUS      { E_Operator Op_MINUS }
  | T_UPLUS;T_AT  { E_UOperator Op_UPlus }
  | T_UMINUS;T_AT { E_UOperator Op_UMinus }
  | T_TILDE;      { E_UOperator Op_UTilde }

assignable:
  | e = identifier  { e }
  | k = keyword_as_id { k }
  | T_LBRACK;T_RBRACK;
  | T_LBRACK_ARG;T_RBRACK; { E_Operator Op_AREF }

method_name:
  | e = message_identifier { e }
  | e = method_name; T_ASSIGN; { add_eq e }
  | id1 = method_name; T_DOT; n; id2 = message_identifier
    { E_Binop(id1,Op_DOT,id2) }
  | id1 = method_name; T_DOT; n;id2 = assignable; T_ASSIGN
    { E_Binop(id1,Op_DOT,add_eq id2) }
  | T_USCOPE; id = identifier;
  | T_USCOPE; id = keyword_as_id;
    { E_Unary(Op_UScope,id)}
  | id1 = method_name; T_SCOPE; id2 = identifier;
  | id1 = method_name; T_SCOPE; id2 = keyword_as_id;
    { E_Binop(id1,Op_SCOPE,id2) }

compound_with_assign:
  | e1 = compound_expr; T_ASSIGN; n; e2 = compound_with_assign
      { E_Binop (e1,Op_ASSIGN,e2) }
  | e = compound_expr { e }

compound_expr:
  | e = obj_expr 
  | e = simple_expr(compound_expr) 
  | e = delim_expr {e}
  | e = star_exp(obj_expr)   { e }
  | e = star_exp(delim_expr)   { e }

hash_literal:
  | T_LBRACE; n; T_RBRACE { E_Hash [] }
  | T_LBRACE; n; hbody=hash_body;T_RBRACE  { E_Hash hbody }

delim_expr:
  | e = method_with_codeblock(delim_expr) { e }
  | e = hash_literal { e }
  | e = paren_exp { e }
  | e = paren_brackets { e }

  | K_WHILE; guard=guard_expr; ctrl_sep(K_DO); body=expression_list; K_lEND;
    { E_While(guard,body) }

  | K_UNTIL; guard=guard_expr; ctrl_sep(K_DO); body=expression_list; K_lEND;
    { E_Until(guard,body) }

  | K_IF; guard = guard_expr; ctrl_sep(K_THEN); es = expression_list;
    else_e = if_else_clauses; K_lEND
    { E_If(guard,es,else_e) }

  | K_UNLESS; guard=guard_expr; ctrl_sep(K_THEN); es = expression_list; 
    else_e = case_else; K_lEND
    { E_Unless(guard,es, else_e) }

return_arg_paren:
  | e = return_arg {e}
  | e = paren_exp {e}
return_arg:
  | e = method_arg { e }
  | e = hash_literal { e }

method_arg:
  | e = obj_expr
  | e = star_exp(obj_expr)
  | e = simple_expr(method_arg) {e}
  | e1 = method_arg; T_ASSIGN; n; e2 = method_arg
      { E_Binop (e1,Op_ASSIGN,e2) }


simple_expr(rec_expr):
  | e1 = rec_expr; op = bin_op_no_assign; n; e2 = compound_expr
      { E_Binop (e1,op,e2) }

  | op = unary_op; e = compound_expr
      { E_Unary(op,e) }

  | e1 = rec_expr; T_QUESTION; n;
    e2 = compound_expr; T_COLON;n; e3 = compound_expr
    { E_Tertiary (e1,e2,e3) }

method_formals:
  | nonempty_list(T_EOL) { [] }
  | T_SEMICOLON; { [] }    
  | formals = separated_nonempty_list(pair(T_COMMA,n), formal_arg); nonempty_list(T_EOL)
    {formals}

  | T_LPAREN; n;formals = separated_list(pair(T_COMMA,n), formal_arg); 
    T_RPAREN;n
  | T_LPAREN_ARG; n;formals = separated_list(pair(T_COMMA,n), formal_arg); 
    T_RPAREN;n
    { formals }

formal_paren:
  | T_LPAREN; e = comma_list_suffix(formal_paren); T_RPAREN {Formal_tuple e}
  | e = formal_arg { e }

formal_arg:
  | id = T_LID  { Formal_id id }
  | T_UAMPER; id = T_LID { Formal_amp id }
  | T_USTAR; id = T_LID  { Formal_star id}
  | T_USTAR {Formal_rest}
  | f = formal_arg; T_ASSIGN; e = compound_expr
   { Formal_default (f,e) }

code_block:
  | K_DO; n; b = code_block_body; K_lEND
  | T_LBRACE; n; b = code_block_body; T_RBRACE
    { let args,body = b in E_CodeBlock(args,body) }

code_block_body:
(*
  | T_VBAR; args = separated_nonempty_list(T_COMMA, formal_arg); T_VBAR;n;
    body = expression_list
*)
  | T_VBAR; args = comma_list_suffix(formal_paren); T_VBAR;n; 
    body = expression_list
    {(args,body)}
  | body = expression_list
    {([], body)}

meth_or_atom:
    | e = method_name  {e}
    | a = T_ATOM { E_Literal (Lit_Atom a) }

scoped(base):
  | e = base { e }
  | T_USCOPE; id = identifier { E_Unary(Op_UScope,id)}
  | id1 = scoped(base); T_SCOPE; id2 = identifier;
  | id1 = scoped(base); T_SCOPE; id2 = keyword_as_id;
    { E_Binop(id1,Op_SCOPE,id2) }

identifier:
  | id = T_BUILTIN_VAR { E_Identifier(ID_Builtin, id) }
  | id = T_GLOBAL_VAR  { E_Identifier (ID_Global, id) }
  | id = raw_id        { id }
  | T_AT; e = keyword_as_id;
  | T_AT; e = raw_id   { make_inst_var e}
  | T_AT;T_AT; e = keyword_as_id;
  | T_AT;T_AT; e = raw_id   { make_class_var e }

raw_id:
  | id = T_LID { E_Identifier (ID_Lowercase, id)}
  | id = T_UID { E_Identifier (ID_Uppercase, id)}
  | K_NIL      { E_Literal Lit_Nil }
(*
  | K_SUPER    { E_Literal Lit_Super }
  | K_SELF     { E_Literal Lit_Self }
  | K_UNDEF    { E_Literal Lit_Undef }*)

constant:
  | s = T_TICK { E_Literal (Lit_String (String_Tick, s)) }
  | s = T_SINGLE_STRING { E_Literal (Lit_String (String_Single, s)) }
  | s = T_DOUBLE_STRING { E_Literal (Lit_String (String_Double, s)) }
  | s = T_USER_STRING 
    { let m,str = s in
	match m with
	  | "r" -> E_Literal(Lit_Regexp (str,"")) 
	  | _ -> E_Literal (Lit_String(String_User, str))
    }
  | i = T_HEREDOC { E_Literal(Lit_String (String_Heredoc i,""))}
  | r = T_REGEXP { let (s,m) = r in E_Literal(Lit_Regexp (s,m)) }
  | i = T_FIXNUM { E_Literal(Lit_FixNum i) }
  | i = T_BIGNUM { E_Literal(Lit_BigNum i) }
  | f = T_FLOAT { E_Literal(Lit_Float f) }
  | a = T_ATOM { E_Literal (Lit_Atom a) }
(*  
  | K_BREAK  { E_Literal Lit_Break }
  | K_YIELD  { E_Literal Lit_Yield }
  | K_REDO   { E_Literal Lit_Redo }
  | K_NEXT   { E_Literal Lit_Next }
  | K_TRUE   { E_Literal Lit_True }
  | K_FALSE  { E_Literal Lit_False }
  | K_RETRY  { E_Literal Lit_Retry }*)

keyword_as_id:
  | K_CLASS   { E_Identifier(ID_Lowercase, "class") }
  | K_MODULE  { E_Identifier(ID_Lowercase, "module") }
  | K_DEF     { E_Identifier(ID_Lowercase, "def") }
  | K_END     { E_Identifier(ID_Uppercase, "END") }
  | K_ALIAS   { E_Identifier(ID_Lowercase, "alias") }
  | K_BEGIN   { E_Identifier(ID_Lowercase, "BEGIN") }
  | K_RESCUE  { E_Identifier(ID_Lowercase, "rescue") }
  | K_ENSURE  { E_Identifier(ID_Lowercase, "ensure") }
  | K_IF      { E_Identifier(ID_Lowercase, "if") }
  | K_UNLESS  { E_Identifier(ID_Lowercase, "unless") }
  | K_THEN    { E_Identifier(ID_Lowercase, "then") }
  | K_ELSIF   { E_Identifier(ID_Lowercase, "elsif") }
  | K_ELSE    { E_Identifier(ID_Lowercase, "else") }
  | K_CASE    { E_Identifier(ID_Lowercase, "case") }
  | K_WHEN    { E_Identifier(ID_Lowercase, "when") }
  | K_WHILE   { E_Identifier(ID_Lowercase, "while") }
  | K_UNTIL   { E_Identifier(ID_Lowercase, "until") }
  | K_FOR     { E_Identifier(ID_Lowercase, "for") }
  | K_IN      { E_Identifier(ID_Lowercase, "in") }
  | K_DO      { E_Identifier(ID_Lowercase, "do") }
  | K_RETURN  { E_Identifier(ID_Lowercase, "return") }
  | K_AND     { E_Identifier(ID_Lowercase, "and") }
  | K_OR      { E_Identifier(ID_Lowercase, "or") }
  | K_NOT     { E_Identifier(ID_Lowercase, "not") }
  | K_DEFINED { E_Identifier(ID_Lowercase, "defined?") }
  | K_lBEGIN  { E_Identifier(ID_Lowercase, "begin") }
  | K_lEND    { E_Identifier(ID_Lowercase, "end") }
  | K_YIELD   { E_Identifier(ID_Lowercase, "yield") }
(*
const_key_as_id:
  | K_BREAK   { E_Identifier(ID_Lowercase, "break") }
  | K_REDO    { E_Identifier(ID_Lowercase, "redo") }
  | K_NEXT    { E_Identifier(ID_Lowercase, "next") }
  | K_RETRY   { E_Identifier(ID_Lowercase, "retry") }
  | K_TRUE    { E_Identifier(ID_Lowercase, "true") }
  | K_FALSE   { E_Identifier(ID_Lowercase, "false") }

  | K_NIL     { E_Identifier(ID_Lowercase, "nil") }
  | K_SELF    { E_Identifier(ID_Lowercase, "self") }
  | K_UNDEF   { E_Identifier(ID_Lowercase, "undef") }
  | K_SUPER   { E_Identifier(ID_Lowercase, "super") }*)
      
when_clauses:
  | { [] }
  | K_WHEN; e=separated_nonempty_list(pair(T_COMMA,n),inner_guard);
    ctrl_sep(K_THEN) es=expression_list; rest=when_clauses
    { (e,es) :: rest }
    
case_else:
  | { [] }
  | K_ELSE;n; es = expression_list { es }

rescue_clause:
  | K_RESCUE; T_SEMICOLON; n; body = expression_list
    { (E_Empty,E_Block body) }

  | K_RESCUE; nonempty_list(T_EOL); body = expression_list
    { (E_Empty,E_Block body) }
  | K_RESCUE; T_ASSOC; bind = identifier;n; body = expression_list
     /* FIXME */
    { (E_Empty,E_Block body) }
  | K_RESCUE; exn = separated_nonempty_list(T_COMMA,inner_guard); n; body = expression_list
    { (E_Tuple exn,E_Block body) }

ensure_clause:
  | { [] }
  | K_ENSURE; n; body = expression_list { body }

if_else_clauses:
  | { [] }
  | K_ELSE;n; es = expression_list
    { es }
  | K_ELSIF;n; guard = guard_expr; ctrl_sep(K_THEN); es = expression_list; 
    rest = if_else_clauses
    { [E_If( guard, es, rest) ] }

%inline unary_op:
  | T_UPLUS   { Op_UPlus }
  | T_UMINUS  { Op_UMinus }
  | T_BANG    { Op_UBang }
  | T_TILDE   { Op_UTilde }
  | K_NOT     { Op_UNot }
  | T_UAMPER  { Op_UAmper }
(*  | K_DEFINED { Op_UDefined}*)
(*  | T_USTAR   { Op_UStar }*)

%inline bin_op:
  | T_ASSIGN { Op_ASSIGN }
  | o = bin_op_no_assign {o}

%inline bin_op_no_assign:
  | T_SLASH  { Op_DIV }   
  | T_PERCENT{ Op_REM }
  | T_CMP    { Op_CMP } 
  | T_EQ     { Op_EQ }
  | T_EQQ    { Op_EQQ }
  | T_NEQ    { Op_NEQ }
  | T_GEQ    { Op_GEQ }
  | T_LEQ    { Op_LEQ }
  | T_ANDOP  { Op_AND }
  | T_OROP   { Op_OR }
  | T_MATCH  { Op_MATCH }
  | T_NMATCH { Op_NMATCH }
  | T_CARROT { Op_XOR }
  | T_POW    { Op_POW }
  | T_VBAR   { Op_BOR }
  | T_ASSOC  { Op_ASSOC }
  | T_AMPER  { Op_BAND }
  | T_PLUS   { Op_PLUS }
  | T_MINUS  { Op_MINUS }
  | T_STAR   { Op_TIMES }
  | T_LSHFT  { Op_LSHIFT }
  | T_RSHFT  { Op_RSHIFT }
  | T_LT     { Op_LT }
  | T_GT     { Op_GT }
  | T_DOT2   { Op_DOT2 }
  | T_DOT3   { Op_DOT3 }
  | o = T_OP_ASGN {Op_OP_ASGN o }


lparen:
  | T_LPAREN_ARG {}
  | T_LPAREN {}
