
open Cfg
open Cfg_printer.CodePrinter
open Utils
open Format
open Config
open Annotation

(*
let rec format_def_name ppf = function
  | Instance_Method e -> format_msg_id ppf e
  | Singleton_Method(e1, e2) -> 
      fprintf ppf "%a.%a" format_identifier e1 format_msg_id e2
*)


let emit_id ppf ident = 
  let name = format_to_string format_type_ident ident in

    if name = "::Boolean"
    then fprintf ppf "Contract::UnionType.new([Contract::Type.new('::TrueClass'),Contract::Type.new('::FalseClass')])"
    else fprintf ppf "Contract::Type.new('%s')" name

let rec emit_type_expr ppf e = match e with
  | Type_Ident ident -> emit_id ppf ident

  | Type_Union lst ->
      fprintf ppf "Contract::UnionType.new([%a])"
        (format_comma_list emit_type_expr) lst
      
  | Type_Varargs _ | Type_Optional _
  | Type_ParamList _
  | Type_Var _
  | Type_App _
  | Type_Object _->
      Log.fixme "contract emit unsupported";
      fprintf ppf "Contract::Type.new('Object')"

  | Type_Dynamic
  | Type_Fixme -> 
      Log.fixme "contract emit unsupported";
      fprintf ppf "Contract::Type.new('Object')"

let is_vararg = function 
  | Type_Varargs _ -> true
  | _ -> false

let is_optional = function 
  | Type_Optional _ -> true
  | _ -> false

let emit_params ppf pl = 
  let emit ppf = function
    | Type_Optional e | Type_Varargs e -> emit_type_expr ppf e
    | e -> emit_type_expr ppf e
  in
  let varargs,rest = List.partition is_vararg pl in
  let opts,reqs = List.partition is_optional rest in
    fprintf ppf "Contract::Params.new([%a]"
      (format_comma_list emit_type_expr) reqs;
    begin match opts with
      | [] -> ()
      | _ -> fprintf ppf ",[%a]" (format_comma_list emit) opts;
    end;
    begin match varargs with
      | [] -> ()
      | [v] -> fprintf ppf ",%a" emit v
      | _ -> assert false
    end;
    fprintf ppf ")"
      
let emit_block ppf (MethodSig(args,blk,ret)) = 
  assert(blk = None);
  fprintf ppf "Contract::Block.new(%a,%a)"
    emit_params args
    emit_type_expr ret

let emit_pos ppf pos = 
  fprintf ppf "Contract::SourcePosition.new('%s',%d)"
    pos.Lexing.pos_fname pos.Lexing.pos_lnum

let emit_block_or_nil ppf = function
  | None -> pp_print_string ppf "nil"
  | Some x -> emit_block ppf x

let emit_method ppf (name,binders,(MethodSig(args,blk,ret))) =
  fprintf ppf "%a,%a,%a"
    emit_params args
    emit_block_or_nil blk
    emit_type_expr ret

let emit_annot_mono pos ppf m =
  fprintf ppf "Contract::MonoMethod.new(%a,%a)\n"
    emit_method m
    emit_pos pos

let emit_annot_inter pos ppf alst = 
  let emit_one ppf m = fprintf ppf "[%a]" emit_method m in
    fprintf ppf "Contract::InterMethod.new([%a],%a)\n"
      (format_comma_list emit_one) alst
      emit_pos pos

let out_ppf_ref = ref None

let open_output_printer () = match !out_ppf_ref with
  | None ->
      begin try 
        let oc = open_out "casts.rb" in
        let ppf = formatter_of_out_channel oc in
          fprintf ppf "require 'cast_insertion'\n\n";
          ppf
      with _ -> err_formatter
      end
  | Some oc -> oc

let emit_method_contract clz annot pos = 
  let ann_lst = match annot with
    | MethodType lst -> lst
    | _ -> Log.fatal (Log.of_loc pos) "wrong kind of annotation for method"
  in
  if conf.emit_runtime_casts then begin
    let out_ppf = open_output_printer() in
    let name = match ann_lst with
      | [] -> assert false
      | (name,_,_)::_ -> format_to_string format_type_ident name
    in
    let clz = 
      if substr "::" clz 
      then String.sub clz 2 (String.length clz - 2)
      else clz
    in
      begin match ann_lst with
        | [] -> assert false
        | [m] -> fprintf out_ppf "sig = %a\n" (emit_annot_mono pos) m
        | lst -> fprintf out_ppf "sig = %a\n" (emit_annot_inter pos) lst
      end;
      fprintf out_ppf "Contract::Registry.register('%s',:'%s',sig)\n\n" clz name
  end

let flush_ppf () = match !out_ppf_ref with
  | None -> ()
  | Some ppf -> pp_print_flush ppf ()

let () = at_exit flush_ppf
               
    
(*
and annot_method = {
  annot_meth_name : def_name;
  annot_meth_binders : TypeAnnot_helper.binder list;
  annot_meth_typ : annot_func;
  annot_meth_block : annot_func option;
  annot_meth_cons : annot_cons list;
}
  ()
*)  
